'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var S = _interopDefault(require('s-js'));

function comparer(v, k, b, isArray, path, r) {
  let index;
  const newPath = path.concat([k]);
  if (isArray && v != null && typeof v === 'object'
      && k != (index = b.findIndex(i => i && (i === v || (v.id != null && i.id === v.id) || (v._id != null && i._id === v._id)))))  {
    return r.push(newPath.concat([index > -1 ? b[index] : v]));
  }
  return r.push.apply(r, diff(v, b[k], newPath));
}

function clone(v) {
  if (!isObject(v)) return v;
  if (Array.isArray(v)) return v.slice(0);
  return Object.assign({}, v);
}

function isObject(obj) {
  let ref;
  return obj !== null && ((ref = typeof obj) === 'object' || ref === 'function');
}

function diff(a, b, path = []) {
  let i, k, l, len, v;
  const r = [];
  if (!isObject(a) || (b == null)) {
    if (a !== b) {
      r.push(path.concat([a]));
    }
  } else if (Array.isArray(a)) {
    for (k = i = 0, len = a.length; i < len; k = ++i) {
      v = a[k];
      if ((b != null ? b[k] : void 0) !== v) comparer(v, k, b, true, path, r);
    }
    if ((b != null ? b.length : void 0) > a.length) {
      l = a.length;
      while (l < b.length) {
        r.push(path.concat([l, void 0]));
        l++;
      }
    }
  } else {
    for (k in a) {
      v = a[k];
      if ((b != null ? b[k] : void 0) !== v)
        comparer(v, k, b, false, path, r);
    }
    for (k in b) {
      v = b[k];
      if (!(k in a))
        r.push(path.concat([k, void 0]));
    }
  }
  return r;
}

function unwrap(item, deep) {
  let result, unwrapped, v;
  if (result = item != null ? item._state : void 0) return result;
  if (!deep || !isObject(item) || (typeof item === 'function') || (item instanceof Element)) return item;
  if (Object.isFrozen(item)) item = clone(item);

  if (Array.isArray(item)) {
    for (let i = 0, l = item.length; i < l; i++) {
      v = item[i];
      if ((unwrapped = unwrap(v, deep)) !== v) item[i] = unwrapped;
    }
  } else {
    let keys = Object.keys(item);
    for (let i = 0, l = keys.length; i < l; i++) {
      v = item[keys[i]];
      if ((unwrapped = unwrap(v, deep)) !== v) item[keys[i]] = unwrapped;
    }
  }
  return item;
}

const SNODE = Symbol('solid-node'),
  SPROXY = Symbol('solid-proxy');

function getDataNode(target) {
  let node = target[SNODE];
  if (!node) target[SNODE] = node = {};
  return node;
}

function trigger(node, property, notify) {
  if (node[property]) node[property].next();
  if (notify && node._self) node._self.next();
}

function track(target, property, value) {
  let node;
  if (isObject(value) && !(value instanceof Element)) {
    if (node = getDataNode(value)) {
      if (!node._self) node._self = S.makeDataNode();
      node._self.current();
    }
  }
  node = getDataNode(target);
  node[property] || (node[property] = S.makeDataNode());
  node[property].current();
}

function setNested(item, changes) {
  let node = getDataNode(item),
    isArray = Array.isArray(item),
    value, notify, keys;

  if (arguments.length === 3) {
    notify = isArray || !(arguments[1] in item);
    value = unwrap(arguments[2], true);
    if (item[arguments[1]] === value) return;
    if (value === void 0) {
      delete item[arguments[1]];
      if (isArray) item.length--;
    } else item[arguments[1]] = value;

    trigger(node, arguments[1], notify);
    return;
  }

  keys = Object.keys(changes);
  for (let i = 0, l = keys.length; i < l; i++) {
    const property = keys[i];
    notify = isArray || !(property in item);
    value = unwrap(changes[property], true);
    if (value === void 0) delete item[property];
    else item[property] = value;
    trigger(node, property, notify);
  }
}

function resolvePath(current, path, length) {
  let i = 0, temp;
  while (i < length && (temp = current[path[i]]) != null) {
    current = temp;
    i++;
  }
  return current;
}

function wrap(value) { return value[SPROXY] || (value[SPROXY] = new Proxy(value, proxyTraps)); }

function resolveAsync(value, fn) {
  if (!isObject(value)) return fn(value);
  if ('subscribe' in value) {
    const dispose = value.subscribe(fn);
    S.cleanup(function disposer() { dispose.unsubscribe(); });
    return;
  }
  if ('then' in value) {
    value.then(fn);
    return;
  }
  fn(value);
}

const proxyTraps = {
  get(target, property) {
    if (property === '_state') return target;
    const value = target[property];
    if (S.isListening() && typeof value !== 'function') track(target, property, value);
    return (!isObject(value) || typeof value === 'function' || value instanceof Element) ? value : wrap(value);
  },

  set() { return true; },

  deleteProperty() { return true; }
};

class State {
  constructor(state = {}) {
    Object.defineProperties(this, {
      _state: { value: unwrap(state, true), writable: true }
    });
    const keys = Object.keys(this._state);
    for (let i = 0, l = keys.length; i < l; i++) this._defineProperty(keys[i]);
  }

  set() {
    const args = arguments;
    S.freeze(() => {
      if (args.length === 1) {
        if (Array.isArray(args[0])) {
          for (let i = 0; i < args[0].length; i++) this.set.apply(this, args[0][i]);
        } else {
          const keys = Object.keys(args[0]);
          for (let i = 0, l = keys.length; i < l; i++) {
            const property = keys[i];
            this._setProperty(property, args[0][property]);
          }
        }
        return;
      }

      setNested(resolvePath(this._state, args, args.length - 1), args[args.length - 1]);
    });
    return this;
  }

  replace() {
    if (arguments.length === 1) {
      if (!(arguments[0] instanceof Object)) {
        console.log('replace must be provided a replacement state');
        return this;
      }
      let changes = arguments[0];
      S.freeze(() => {
        if (!(Array.isArray(changes))) changes = diff(changes, this._state);

        for (let i = 0; i < changes.length; i++) this.replace.apply(this, changes[i]);
      });
      return this;
    }

    if (arguments.length === 2) {
      this._setProperty.apply(this, arguments);
      return this;
    }

    const value = arguments[arguments.length - 1],
      property = arguments[arguments.length - 2];
    setNested(resolvePath(this._state, arguments, arguments.length - 2), property, value);
    return this;
  }

  select() {
    const mapFn1 = selection => () => {
      const unwrapped = unwrap(selection(), true),
        results = [];
      resolveAsync(unwrapped, (value) => {
        if (value === void 0) return;
        for (let key in value || {}) {
          results.push(diff(value[key], this._state[key], [key]));
        }
        this.replace([].concat(...results));
      });
    };

    const mapFn2 = (key, selector) => () => {
      const unwrapped = unwrap(selector(), true);
      resolveAsync(unwrapped, (value) => {
        if (value === void 0) return;
        this.replace(diff(value, this._state[key], [key]));
      });
    };

    for (let i = 0; i < arguments.length; i++) {
      const selection = arguments[i];
      if (typeof selection === 'function') {
        S.makeComputationNode(mapFn1(selection));
        continue;
      }
      for (let key in selection) {
        if (!(key in this)) this._defineProperty(key);
        S.makeComputationNode(mapFn2(key, selection[key]));
      }
    }
    return this;
  }

  _setProperty(property, value) {
    if (!(property in this)) this._defineProperty(property);
    value = unwrap(value, true);
    if (this._state[property] === value) return;
    if (value === void 0) delete this._state[property];
    else this._state[property] = value;
    trigger(getDataNode(this._state), property);
  }

  _defineProperty(property) {
    Object.defineProperty(this, property, {
      get() {
        const value = this._state[property];
        if (S.isListening()) track(this._state, property, value);
        return (!isObject(value) || typeof value === 'function' || value instanceof Element) ? value : wrap(value);
      },
      enumerable: true
    });
  }
}

function fromPromise(promise, seed) {
  let s = S.makeDataNode(seed),
    complete = false;
  promise
    .then((value) => {
      if (complete) return;
      s.next(value);
    }).catch(err => console.error(err));

  S.cleanup(function dispose() { complete = true; });
  return () => s.current();
}

function fromObservable(observable, seed) {
  let s = S.makeDataNode(seed),
    disposable = observable.subscribe(v => s.next(v), err => console.error(err));

  S.cleanup(function dispose() {
    disposable.unsubscribe();
    disposable = null;
  });
  return () => s.current();
}

function from(input, seed) {
  if (isObject(input)) {
    if (typeof input === 'function') return input;
    if (Symbol.observable in input) return fromObservable(input[Symbol.observable](), seed);
    if ('then' in input) return fromPromise(input, seed);
  }
  throw new Error('from() input must be a function, Promise, or Observable');
}

function pipe(input, ...fns) {
  return compose(...fns)(input);
}

function map(fn) {
  return input => () => {
    const value = input();
    if (value === void 0) return;
    return S.sample(() => fn(value));
  }
}

function compose(...fns) {
  if (!fns) return i => i;
  if (fns.length === 1) return fns[0];
  return input => fns.reduce(((prev, fn) => fn(prev)), input);
}

// memoized map that handles falsey rejection
function when(mapFn) {
  let mapped, value, disposable;
  S.cleanup(function dispose() {
    disposable && disposable();
  });
  return map(function mapper(newValue) {
    if (newValue == null || newValue === false) {
      disposable && disposable();
      return value = mapped = disposable = null;
    }
    if (value === newValue) return mapped;
    disposable && disposable();
    disposable = null;
    value = newValue;
    return mapped = S.root((d) => {
      disposable = d;
      return mapFn(value);
    });
  })
}

// Need to be able grab wrapped state internals so can't use S-Array
function each(mapFn) {
  let mapped = [],
      list = [],
      disposables = [],
      length = 0;
  S.cleanup(function dispose() {
    for (let i = 0; i < disposables.length; i++) disposables[i]();
  });
  return map(function mapper(newList) {
    let i, j = 0,
      newLength = (newList && newList.length) || 0;
    if (newLength === 0) {
      if (length !== 0) {
        for (i = 0; i < length; i++) disposables[i]();
        list = [];
        mapped = [];
        disposables = [];
        length = 0;
      }
    } else if (length === 0) {
      j = 0;
      while (j < newLength) {
        list[j] = newList[j];
        mapped[j] = S.root(mappedFn);
        j++;
      }
      length = newLength;
    } else {
      const newMapped = new Array(newLength),
        tempDisposables = new Array(newLength),
        indexedItems = new Map();

      // reduce from both ends
      let end = Math.min(length, newLength),
        start = 0, item, itemIndex, newEnd;
      while (start < end && newList[start] === list[start]) start++;

      end = length - 1;
      newEnd = newLength - 1;
      while (end >= 0 && newEnd >= 0 && newList[newEnd] === list[end]) {
        newMapped[newEnd] = mapped[end];
        tempDisposables[newEnd] = disposables[end];
        end--;
        newEnd--;
      }

      // create indices
      j = newEnd;
      while (j >= start) {
        item = newList[j];
        itemIndex = indexedItems.get(item);
        if (itemIndex != null) itemIndex.push(j);
        else indexedItems.set(item, [j]);
        j--;
      }

      // find old items
      i = start;
      while (i <= end) {
        item = list[i];
        itemIndex = indexedItems.get(item);
        if (itemIndex != null && itemIndex.length > 0) {
          j = itemIndex.pop();
          newMapped[j] = mapped[i];
          tempDisposables[j] = disposables[i];
        } else disposables[i]();
        i++;
      }

      // set all new values
      j = start;
      while (j < newLength) {
        list[j] = newList[j];
        if (newMapped.hasOwnProperty(j)) {
          mapped[j] = newMapped[j];
          disposables[j] = tempDisposables[j];
        } else mapped[j] = S.root(mappedFn);
        j++;
      }

      // truncate extra length
      length = list.length = mapped.length = disposables.length = newLength;
    }
    return mapped;

    function mappedFn(dispose) {
      disposables[j] = dispose;
      return mapFn(list[j], j);
    }
  });
}

// export observable
function observable(input) {
  if (Symbol.observable in input) return input[Symbol.observable]();
  return {
    subscribe(observer) {
      if (!(observer instanceof Object) || observer == null) {
        throw new TypeError('Expected the observer to be an object.');
      }
      observer = observer.next || observer;
      let complete = false;
      S.on(input, function next() {
        if (complete) return;
        observer(input());
      });
      return {
        unsubscribe() { complete = true; }
      };
    },
    [Symbol.observable]() { return this; }
  };
}

const { root, cleanup, sample, data, effect } = S;

exports.root = root;
exports.cleanup = cleanup;
exports.sample = sample;
exports.data = data;
exports.effect = effect;
exports.State = State;
exports.unwrap = unwrap;
exports.from = from;
exports.pipe = pipe;
exports.map = map;
exports.compose = compose;
exports.when = when;
exports.each = each;
exports.observable = observable;
